﻿// "Therefore those skilled at the unorthodox
// are infinite as heaven and earth,
// inexhaustible as the great rivers.
// When they come to an end,
// they bagin again,
// like the days and months;
// they die and are reborn,
// like the four seasons."
// 
// - Sun Tsu,
// "The Art of War"

using System.Drawing;
using HtmlRenderer.Entities;
using HtmlRenderer.Utils;
using System.Linq;

namespace HtmlRenderer.Dom
{
    /// <summary>
    /// CSS box for image element.
    /// </summary>
    internal sealed class CssBoxImage : CssBox
    {
        #region Fields and Consts

        /// <summary>
        /// the image word of this image box
        /// </summary>
        private readonly CssBoxWord _imageWord;

        /// <summary>
        /// the image rectange restriction as returned from image load event
        /// </summary>
        private Rectangle _imageRectangle;

        /// <summary>
        /// handler used for image loading by source
        /// </summary>
        private ImageLoadHandler _imageLoadHandler;

        /// <summary>
        /// is re-layout is required after image is loaded (if loaded async)<br/>
        /// required if image size is not set directly by the box.
        /// </summary>
        private bool _requireLayoutAfterLoad;

        /// <summary>
        /// is image load is finished, used to know if no image is found
        /// </summary>
        private bool _imageLoadingComplete;

        #endregion


        /// <summary>
        /// Init.
        /// </summary>
        /// <param name="parent">the parent box of this box</param>
        /// <param name="tag">the html tag data of this box</param>
        public CssBoxImage(CssBox parent, HtmlTag tag) 
            : base(parent, tag)
        {
            _imageWord = new CssBoxWord(this, null);
            Words.Add(_imageWord);
        }

        /// <summary>
        /// Get the image of this image box.
        /// </summary>
        public Image Image
        {
            get { return _imageWord.Image; }
        }

        /// <summary>
        /// Paints the fragment
        /// </summary>
        /// <param name="g">the device to draw to</param>
        protected override void PaintImp(Graphics g)
        {
            if (Display != CssConstants.None)
            {
                RectangleF rects = Rectangles.Values.First();
                
                PointF offset = HtmlContainer != null ? HtmlContainer.ScrollOffset : PointF.Empty;
                rects.Offset(offset);

                PaintBackground(g, rects, true, true);
                PaintBorder(g, rects, true, true);

                var word = Words[0];
                RectangleF r = word.Bounds;
                r.Offset(offset);
                r.Height -= ActualBorderTopWidth + ActualBorderBottomWidth + ActualPaddingTop + ActualPaddingBottom;
                r.Y += ActualBorderTopWidth + ActualPaddingTop;

                if (word.Image != null)
                {
                    if (_imageRectangle == Rectangle.Empty)
                        g.DrawImage(word.Image, Rectangle.Round(r));
                    else
                        g.DrawImage(word.Image, Rectangle.Round(r), _imageRectangle, GraphicsUnit.Point);

                    if (word.Selected)
                    {
                        g.FillRectangle(CssUtils.SelectionBackcolor, word.Left - word.LastMeasureOffset.X + offset.X, word.Top + offset.Y, word.Width, DomUtils.GetCssLineBoxByWord(word).LineHeight);
                    }
                }
                else if (_imageLoadingComplete && r.Width > 20 && r.Height > 20)
                {
                    DrawingUtils.DrawImageErrorIcon(g, r);
                }
                else
                {
                    DrawingUtils.DrawImageLoadingIcon(g, r);
                }
            }
        }

        /// <summary>
        /// Assigns words its width and height
        /// </summary>
        /// <param name="g">the device to use</param>
        internal override void MeasureWordsSize(Graphics g)
        {
            if (!_wordsSizeMeasured)
            {
                if(_imageLoadHandler == null)
                {
                    _imageLoadHandler = new ImageLoadHandler(OnLoadImageComplete);
                    _imageLoadHandler.LoadImage(HtmlContainer, GetAttribute("src"), HtmlTag.Attributes);
                }

                MeasureWordSpacing(g); // atodo: why the fuck this effects text after the image?
                MeasureImageSize(_imageWord);
                _wordsSizeMeasured = true;
            }
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public override void Dispose()
        {
            if(_imageLoadHandler != null)
                _imageLoadHandler.Dispose();
            base.Dispose();
        }


        #region Private methods

        /// <summary>
        /// Measure image box size by the width\height set on the box and the actual rendered image size.<br/>
        /// If no image exists for the box error icon will be set.
        /// </summary>
        /// <param name="imageWord">the image word to measure</param>
        private void MeasureImageSize(CssBoxWord imageWord)
        {
            var w = new CssLength(Width);
            var h = new CssLength(Height);

            bool hasImageTagWidth = w.Number > 0 && w.Unit == CssUnit.Pixels;
            bool hasImageTagHeight = h.Number > 0 && h.Unit == CssUnit.Pixels;
            _requireLayoutAfterLoad = !hasImageTagWidth || !hasImageTagHeight;

            if (hasImageTagWidth)
            {
                imageWord.Width = w.Number;
            }
            else if (imageWord.Image != null)
            {
                imageWord.Width = _imageRectangle == Rectangle.Empty ? imageWord.Image.Width : _imageRectangle.Width;
            }
            else
            {
                imageWord.Width = hasImageTagHeight ? h.Number/1.14f : 20;
            }

            if (hasImageTagHeight)
            {
                imageWord.Height = h.Number;
            }
            else if (imageWord.Image != null)
            {
                imageWord.Height = _imageRectangle == Rectangle.Empty ? imageWord.Image.Height : _imageRectangle.Height;
            }
            else
            {
                imageWord.Height = hasImageTagWidth ? w.Number*1.14f : 22.8f;
            }

            if (imageWord.Image != null)
            {
                // If only the width was set in the html tag, ratio the height.
                if (hasImageTagWidth && !hasImageTagHeight)
                {
                    // Devide the given tag width with the actual image width, to get the ratio.
                    float ratio = imageWord.Width/imageWord.Image.Width;
                    imageWord.Height = imageWord.Image.Height*ratio;
                }
                    // If only the height was set in the html tag, ratio the width.
                else if (hasImageTagHeight && !hasImageTagWidth)
                {
                    // Devide the given tag height with the actual image height, to get the ratio.
                    float ratio = imageWord.Height/imageWord.Image.Height;
                    imageWord.Width = imageWord.Image.Width*ratio;
                }
            }
            else if (_imageLoadingComplete)
            {
                SetErrorBorder();
            }

            imageWord.Height += ActualBorderBottomWidth + ActualBorderTopWidth + ActualPaddingTop + ActualPaddingBottom;
        }

        /// <summary>
        /// Set error image border on the image box.
        /// </summary>
        private void SetErrorBorder()
        {
            SetAllBorders(CssConstants.Solid, "1px", "#A0A0A0");
            BorderRightColor = BorderBottomColor = "#E3E3E3";
        }

        /// <summary>
        /// On image load process is complete with image or without update the image box.
        /// </summary>
        /// <param name="image">the image loaded or null if failed</param>
        /// <param name="rectangle">the source rectangle to draw in the image (empty - draw everything)</param>
        /// <param name="async">is the callback was called async to load image call</param>
        private void OnLoadImageComplete(Image image, Rectangle rectangle, bool async)
        {
            _imageWord.Image = image;
            _imageRectangle = rectangle;
            _imageLoadingComplete = true;
            _wordsSizeMeasured = false;

            if (_imageLoadingComplete && Image == null)
                SetErrorBorder();

            if (async)
                HtmlContainer.RequestRefresh(_requireLayoutAfterLoad);
        }

        #endregion
    }
}